from hub.core.index import merge_slices
from pytest_cases import parametrize_with_cases  # type: ignore


class MergeSlicesCases:
    def case_first_longer(self):
        return slice(0, 4), slice(1, 3)

    def case_second_longer(self):
        return slice(0, 4), slice(1, 7)

    def case_start_none(self):
        return slice(None, 4), slice(1, 7)

    def case_stop_first_none(self):
        return slice(1, None), slice(3, 7)

    def case_stop_second_none(self):
        return slice(2, 5), slice(1, None)

    def case_stop_both_none(self):
        return slice(1, None), slice(2, None)

    def case_irregular_steps(self):
        return slice(4, 101, 2), slice(10, 20, 3)

    def case_negative_step(self):
        return slice(None), slice(None, None, -3)


@parametrize_with_cases("first,second", cases=MergeSlicesCases)
def test_merge_slices(first: slice, second: slice):
    r = range(100)
    assert r[first][second] == r[merge_slices(first, second)]



from hub.core.dataset import Dataset
from hub.tests.common import requires_tensorflow
import numpy as np
import hub
import pytest


@requires_tensorflow
def test_tensorflow_with_compression(local_ds: Dataset):
    # TODO: when chunk-wise compression is done, `labels` should be compressed using lz4, so this test needs to be updated
    images = local_ds.create_tensor("images", htype="image", sample_compression="png")
    labels = local_ds.create_tensor("labels", htype="class_label")

    images.extend(np.ones((16, 10, 10, 3), dtype="uint8"))
    labels.extend(np.ones((16, 1), dtype="uint32"))

    for batch in local_ds.tensorflow():
        # converting tf Tensors to numpy
        X = batch["images"].numpy()
        T = batch["labels"].numpy()
        assert X.shape == (10, 10, 3)
        assert T.shape == (1,)


@requires_tensorflow
def test_tensorflow_small(local_ds):
    local_ds.create_tensor("image")
    local_ds.image.extend(np.array([i * np.ones((10, 10)) for i in range(256)]))
    local_ds.create_tensor("image2")
    local_ds.image2.extend(np.array([i * np.ones((12, 12)) for i in range(256)]))
    local_ds.flush()

    tds = local_ds.tensorflow()
    for i, batch in enumerate(tds):
        # converting tf Tensors to numpy
        np.testing.assert_array_equal(batch["image"].numpy(), i * np.ones((10, 10)))
        np.testing.assert_array_equal(batch["image2"].numpy(), i * np.ones((12, 12)))


@requires_tensorflow
def test_corrupt_dataset(local_ds, corrupt_image_paths, compressed_image_paths):
    img_good = hub.read(compressed_image_paths["jpeg"][0])
    img_bad = hub.read(corrupt_image_paths["jpeg"])
    with local_ds:
        local_ds.create_tensor("image", htype="image", sample_compression="jpeg")
        for i in range(3):
            for i in range(10):
                local_ds.image.append(img_good)
            local_ds.image.append(img_bad)
    num_samples = 0
    with pytest.warns(UserWarning):
        tds = local_ds.tensorflow()
        for batch in tds:
            num_samples += 1  # batch_size = 1
    assert num_samples == 30


@requires_tensorflow
def test_groups(local_ds, compressed_image_paths):
    img1 = hub.read(compressed_image_paths["jpeg"][0])
    img2 = hub.read(compressed_image_paths["png"][0])
    with local_ds:
        local_ds.create_tensor(
            "images/jpegs/cats", htype="image", sample_compression="jpeg"
        )
        local_ds.create_tensor(
            "images/pngs/flowers", htype="image", sample_compression="png"
        )
        for _ in range(10):
            local_ds.images.jpegs.cats.append(img1)
            local_ds.images.pngs.flowers.append(img2)
    tds = local_ds.tensorflow()
    for batch in tds:
        np.testing.assert_array_equal(batch["images/jpegs/cats"].numpy(), img1.array)
        np.testing.assert_array_equal(batch["images/pngs/flowers"].numpy(), img2.array)

    with local_ds:
        local_ds.create_tensor(
            "arrays/x",
        )
        local_ds.create_tensor(
            "arrays/y",
        )
        for _ in range(10):
            local_ds.arrays.x.append(np.random.random((2, 3)))
            local_ds.arrays.y.append(np.random.random((4, 5)))

    tds = local_ds.images.tensorflow()
    for batch in tds:
        np.testing.assert_array_equal(batch["jpegs/cats"].numpy(), img1.array)
        np.testing.assert_array_equal(batch["pngs/flowers"].numpy(), img2.array)


@requires_tensorflow
def test_tensorflow_string_objects(local_ds: Dataset):
    with local_ds:
        local_ds.create_tensor("strings", htype="text")
        local_ds.strings.extend([f"testing{idx}" for idx in range(10)])  # type: ignore[attr-defined]
    tds = local_ds.tensorflow()
    for idx, batch in tds.enumerate():
        np.testing.assert_array_equal(
            batch["strings"].numpy(), [f"testing{idx}".encode()]
        )



import pytest
import platform
from hub.constants import MB
from hub.util.check_installation import pytorch_installed

if not pytorch_installed():
    pytest.skip("pytroch is not installed", allow_module_level=True)

if platform.system() in ["Windows", "Darwin"]:
    pytest.skip("mock pickling gets quite messy on win/mac", allow_module_level=True)

from unittest.mock import patch

from hub.core.io import SampleStreaming, IOBlock, Schedule
from hub.integrations.pytorch.dataset import (
    ShufflingIterableDataset,
    SubIterableDataset,
)
from hub.integrations.pytorch.common import (
    collate_fn as default_collate_fn,
)

import torch
from torch.utils.data.dataloader import DataLoader
import numpy


def list_blocks(streaming=None):
    return [
        IOBlock(chunks=["a", "b"], indexes=[1, 2, 3]),
        IOBlock(chunks=["c", "b"], indexes=[4, 5, 6]),
        IOBlock(chunks=["d", "e"], indexes=[7]),
    ]


def emit_samples(streaming, schedule: Schedule):
    for block in schedule:
        for i in block.indices():
            yield {"images": numpy.ones((5)) * i}


def throws_exception(streaming, schedule: Schedule):
    yield from emit_samples(streaming, schedule)
    raise RuntimeError("test error")


def mock_dataset(cls):
    instance = cls()
    return instance


@patch.object(SampleStreaming, "list_blocks", list_blocks)
@patch.object(SampleStreaming, "read", emit_samples)
@patch("hub.core.dataset.Dataset")
def test_dataloader(ds):
    dataset = ShufflingIterableDataset(
        mock_dataset(ds), use_local_cache=False, num_workers=2
    )
    dataloader = DataLoader(dataset, batch_size=1, collate_fn=default_collate_fn)

    result = list(dataloader)

    assert len(result) == sum([len(block) for block in list_blocks()])


@patch.object(SampleStreaming, "list_blocks", list_blocks)
@patch.object(SampleStreaming, "read", emit_samples)
@patch("hub.core.dataset.Dataset")
def test_dataloader_batching(ds):
    dataset = ShufflingIterableDataset(
        mock_dataset(ds), use_local_cache=False, num_workers=2
    )
    dataloader = DataLoader(dataset, batch_size=2, collate_fn=default_collate_fn)

    result = list(dataloader)

    assert len(result) == 4


@patch.object(SampleStreaming, "list_blocks", list_blocks)
@patch.object(SampleStreaming, "read", emit_samples)
@patch("hub.core.dataset.Dataset")
def test_more_workers_than_chunk(ds):
    dataset = ShufflingIterableDataset(
        mock_dataset(ds), use_local_cache=False, num_workers=4
    )
    dataloader = DataLoader(dataset, batch_size=1, collate_fn=default_collate_fn)

    result = list(dataloader)

    assert len(result) == sum([len(block) for block in list_blocks()])


@patch.object(SampleStreaming, "list_blocks", list_blocks)
@patch.object(SampleStreaming, "read", emit_samples)
@patch("hub.core.dataset.Dataset")
def test_big_buffer_size(ds):
    dataset = ShufflingIterableDataset(
        mock_dataset(ds), use_local_cache=False, num_workers=4, buffer_size=512
    )
    dataloader = DataLoader(dataset, batch_size=1, collate_fn=default_collate_fn)

    result = list(dataloader)

    assert len(result) == sum([len(block) for block in list_blocks()])


def mock_tranform_f(data):
    return {"items": torch.tensor(42)}


@patch.object(SampleStreaming, "list_blocks", list_blocks)
@patch.object(SampleStreaming, "read", emit_samples)
@patch("hub.core.dataset.Dataset")
def test_workers_transform(ds):
    dataset = ShufflingIterableDataset(
        mock_dataset(ds),
        use_local_cache=False,
        num_workers=4,
        transform=mock_tranform_f,
    )
    dataloader = DataLoader(dataset, batch_size=1, collate_fn=default_collate_fn)

    result = list(dataloader)

    for x in result:
        assert x["items"] == 42


@patch.object(SampleStreaming, "list_blocks", list_blocks)
@patch.object(SampleStreaming, "read", throws_exception)
@patch("hub.core.dataset.Dataset")
def test_proppagete_exception(ds):
    dataset = ShufflingIterableDataset(
        mock_dataset(ds), use_local_cache=False, num_workers=1
    )
    dataloader = DataLoader(dataset, batch_size=1, collate_fn=default_collate_fn)

    with pytest.raises(RuntimeError):
        list(dataloader)


@patch.object(SampleStreaming, "list_blocks", list_blocks)
@patch.object(SampleStreaming, "read", emit_samples)
@patch("hub.core.dataset.Dataset")
def test_method2(ds):
    dataset = SubIterableDataset(
        mock_dataset(ds),
        use_local_cache=False,
        num_workers=2,
        batch_size=2,
        collate_fn=default_collate_fn,
    )
    dataloader = DataLoader(dataset, batch_size=1, collate_fn=default_collate_fn)

    result = list(dataloader)

    assert len(result) == sum([len(block) for block in list_blocks()])



import json
from hub.tests.path_fixtures import gcs_creds
from hub.tests.common import is_opt_true
from hub.tests.storage_fixtures import enabled_storages, enabled_persistent_storages
from hub.tests.cache_fixtures import enabled_cache_chains
from hub.core.storage.gcs import GCloudCredentials
from hub.util.exceptions import GCSDefaultCredsNotFoundError
import os
import pytest
from hub.constants import MB, GCS_OPT
import pickle


KEY = "file"


# helper functions for tests
def check_storage_provider(storage):
    FILE_1 = f"{KEY}_1"
    FILE_2 = f"{KEY}_2"

    storage[FILE_1] = b"hello world"
    assert storage[FILE_1] == b"hello world"
    assert storage.get_bytes(FILE_1, 2, 5) == b"llo"

    storage.set_bytes(FILE_1, b"abcde", 6)
    assert storage[FILE_1] == b"hello abcde"

    storage.set_bytes(FILE_1, b"tuvwxyz", 6)
    assert storage[FILE_1] == b"hello tuvwxyz"

    storage.set_bytes(FILE_2, b"hello world", 3)
    assert storage[FILE_2] == b"\x00\x00\x00hello world"
    storage.set_bytes(FILE_2, b"new_text", overwrite=True)
    assert storage[FILE_2] == b"new_text"

    assert len(storage) >= 1

    for _ in storage:
        pass

    del storage[FILE_1]
    del storage[FILE_2]

    with pytest.raises(KeyError):
        storage[FILE_1]
    storage.flush()


def check_cache_state(cache, expected_state):
    assert cache.dirty_keys == expected_state[0]
    assert set(cache.lru_sizes.keys()) == expected_state[1]
    assert len(cache.cache_storage) == expected_state[2]
    assert len(cache.next_storage) == expected_state[3]
    assert cache.cache_used == expected_state[4]
    assert len(cache) == expected_state[5]


def check_cache(cache):
    chunk = b"0123456789123456" * MB
    FILE_1, FILE_2, FILE_3 = f"{KEY}_1", f"{KEY}_2", f"{KEY}_3"
    check_cache_state(cache, expected_state=[set(), set(), 0, 0, 0, 0])

    cache[FILE_1] = chunk
    check_cache_state(cache, expected_state=[{FILE_1}, {FILE_1}, 1, 0, 16 * MB, 1])

    cache[FILE_2] = chunk
    check_cache_state(
        cache, expected_state=[{FILE_1, FILE_2}, {FILE_1, FILE_2}, 2, 0, 32 * MB, 2]
    )

    cache[FILE_3] = chunk
    check_cache_state(
        cache, expected_state=[{FILE_3, FILE_2}, {FILE_3, FILE_2}, 2, 1, 32 * MB, 3]
    )

    cache[FILE_1]
    check_cache_state(
        cache, expected_state=[{FILE_3}, {FILE_1, FILE_3}, 2, 2, 32 * MB, 3]
    )

    cache[FILE_3]
    check_cache_state(
        cache, expected_state=[{FILE_3}, {FILE_1, FILE_3}, 2, 2, 32 * MB, 3]
    )

    del cache[FILE_3]
    check_cache_state(cache, expected_state=[set(), {FILE_1}, 1, 2, 16 * MB, 2])

    del cache[FILE_1]
    check_cache_state(cache, expected_state=[set(), set(), 0, 1, 0, 1])

    del cache[FILE_2]
    check_cache_state(cache, expected_state=[set(), set(), 0, 0, 0, 0])

    with pytest.raises(KeyError):
        cache[FILE_1]

    cache[FILE_1] = chunk
    check_cache_state(cache, expected_state=[{FILE_1}, {FILE_1}, 1, 0, 16 * MB, 1])

    cache[FILE_2] = chunk
    check_cache_state(
        cache, expected_state=[{FILE_1, FILE_2}, {FILE_1, FILE_2}, 2, 0, 32 * MB, 2]
    )

    cache.flush()
    check_cache_state(cache, expected_state=[set(), {FILE_1, FILE_2}, 2, 2, 32 * MB, 2])

    cache.clear()
    check_cache_state(cache, expected_state=[set(), set(), 0, 0, 0, 0])

    cache[FILE_1] = chunk
    cache[FILE_2] = chunk
    cache.clear_cache()
    check_cache_state(cache, expected_state=[set(), set(), 0, 2, 0, 2])

    cache.clear()
    check_cache_state(cache, expected_state=[set(), set(), 0, 0, 0, 0])


@enabled_storages
def test_storage_provider(storage):
    check_storage_provider(storage)


@enabled_cache_chains
def test_cache(cache_chain):
    check_storage_provider(cache_chain)
    check_cache(cache_chain)


@enabled_persistent_storages
def test_pickling(storage):
    FILE_1 = f"{KEY}_1"
    storage[FILE_1] = b"hello world"
    assert storage[FILE_1] == b"hello world"
    pickled_storage = pickle.dumps(storage)
    unpickled_storage = pickle.loads(pickled_storage)
    assert unpickled_storage[FILE_1] == b"hello world"


@pytest.fixture
def test_gcs_tokens(request):
    if not is_opt_true(request, GCS_OPT):
        pytest.skip()
        return
    gcreds = GCloudCredentials()
    assert gcreds.credentials
    token_path = os.environ["GOOGLE_APPLICATION_CREDENTIALS"]
    gcreds = GCloudCredentials(token=token_path)
    assert gcreds.credentials

    with open(token_path, "rb") as f:
        token = json.load(f)
    gcreds = GCloudCredentials(token=token)
    assert gcreds.credentials
    gcreds = GCloudCredentials(token="google_default")
    assert gcreds.credentials
    gcreds = GCloudCredentials(token="anon")
    assert not gcreds.credentials
    with pytest.raises(GCSDefaultCredsNotFoundError) as e:
        gcreds = GCloudCredentials(token="browser")



import pytest
from hub.util.check_installation import pytorch_installed

if not pytorch_installed():
    pytest.skip("pytroch is not installed", allow_module_level=True)

from hub.integrations.pytorch.shuffle_buffer import ShuffleBuffer
import torch


def test_zero_buffer_size():
    with pytest.raises(ValueError):
        ShuffleBuffer(0)


def test_too_small_buffer():
    buffer = ShuffleBuffer(10)

    tensor = {"val": torch.ones(10)}

    with pytest.warns(UserWarning):
        result = buffer.exchange(tensor)

    assert result == tensor


def test_adding_tensor():
    buffer = ShuffleBuffer(40)
    tensor = {"val": torch.ones(10)}

    result = buffer.exchange(tensor)

    assert result == None


def test_constant_tensor():
    buffer = ShuffleBuffer(8)
    tensor = {"val": torch.tensor(1)}

    result = buffer.exchange(tensor)
    result = buffer.exchange(tensor)

    assert result == tensor



import hub
import numpy as np
import pickle
import pytest

from hub.util.remove_cache import get_base_storage
from hub.util.exceptions import DatasetUnsupportedPytorch, TensorDoesNotExistError
from hub.tests.common import requires_torch
from hub.util.storage import get_pytorch_local_storage
from hub.core.dataset import Dataset
from hub.core.storage.memory import MemoryProvider
from hub.constants import KB

from hub.integrations.pytorch.pytorch_old import dataset_to_pytorch
from hub.tests.dataset_fixtures import enabled_datasets


# ensure tests have multiple chunks without a ton of data
PYTORCH_TESTS_MAX_CHUNK_SIZE = 5 * KB


def to_tuple(sample):
    return sample["image"], sample["image2"]


def pytorch_small_shuffle_helper(start, end, dataloader):
    for _ in range(2):
        all_values = []
        for i, batch in enumerate(dataloader):
            value = batch["image"].numpy()[0][0][0]
            value2 = batch["image2"].numpy()[0][0][0]
            assert value == value2
            all_values.append(value)
            np.testing.assert_array_equal(
                batch["image"].numpy(), value * np.ones(batch["image"].shape)
            )
            np.testing.assert_array_equal(
                batch["image2"].numpy(), value2 * np.ones((1, 12, 12))
            )

    assert set(all_values) == set(range(start, end))


@requires_torch
@enabled_datasets
def test_pytorch_small(ds):
    with ds:
        ds.create_tensor("image", max_chunk_size=PYTORCH_TESTS_MAX_CHUNK_SIZE)
        ds.image.extend(([i * np.ones((i + 1, i + 1)) for i in range(16)]))
        ds.create_tensor("image2", max_chunk_size=PYTORCH_TESTS_MAX_CHUNK_SIZE)
        ds.image2.extend(np.array([i * np.ones((12, 12)) for i in range(16)]))

    if isinstance(get_base_storage(ds.storage), MemoryProvider):
        with pytest.raises(DatasetUnsupportedPytorch):
            dl = ds.pytorch(num_workers=0)
        return

    dl = ds.pytorch(num_workers=2, batch_size=1)

    assert len(dl.dataset) == 16

    for _ in range(2):
        for i, batch in enumerate(dl):
            np.testing.assert_array_equal(
                batch["image"].numpy(), i * np.ones((1, i + 1, i + 1))
            )
            np.testing.assert_array_equal(
                batch["image2"].numpy(), i * np.ones((1, 12, 12))
            )

    dls = ds.pytorch(num_workers=2, batch_size=1, shuffle=True)
    pytorch_small_shuffle_helper(0, 16, dls)

    sub_ds = ds[5:]

    sub_dl = sub_ds.pytorch(num_workers=0)

    for i, batch in enumerate(sub_dl):
        np.testing.assert_array_equal(
            batch["image"].numpy(), (5 + i) * np.ones((1, 6 + i, 6 + i))
        )
        np.testing.assert_array_equal(
            batch["image2"].numpy(), (5 + i) * np.ones((1, 12, 12))
        )

    sub_dls = sub_ds.pytorch(num_workers=0, batch_size=1, shuffle=True)
    pytorch_small_shuffle_helper(5, 16, sub_dls)

    sub_ds2 = ds[8:12]

    sub_dl2 = sub_ds2.pytorch(num_workers=0, batch_size=1)

    for _ in range(2):
        for i, batch in enumerate(sub_dl2):
            np.testing.assert_array_equal(
                batch["image"].numpy(), (8 + i) * np.ones((1, 9 + i, 9 + i))
            )
            np.testing.assert_array_equal(
                batch["image2"].numpy(), (8 + i) * np.ones((1, 12, 12))
            )

    sub_dls2 = sub_ds2.pytorch(num_workers=2, batch_size=1, shuffle=True)
    pytorch_small_shuffle_helper(8, 12, sub_dls2)

    sub_ds3 = ds[:5]

    sub_dl3 = sub_ds3.pytorch(num_workers=0, batch_size=1)

    for _ in range(2):
        for i, batch in enumerate(sub_dl3):
            np.testing.assert_array_equal(
                batch["image"].numpy(), (i) * np.ones((1, i + 1, i + 1))
            )
            np.testing.assert_array_equal(
                batch["image2"].numpy(), (i) * np.ones((1, 12, 12))
            )

    sub_dls3 = sub_ds3.pytorch(num_workers=0, batch_size=1, shuffle=True)
    pytorch_small_shuffle_helper(0, 5, sub_dls3)


@requires_torch
@enabled_datasets
def test_pytorch_transform(ds):
    with ds:
        ds.create_tensor("image", max_chunk_size=PYTORCH_TESTS_MAX_CHUNK_SIZE)
        ds.image.extend(([i * np.ones((i + 1, i + 1)) for i in range(16)]))
        ds.create_tensor("image2", max_chunk_size=PYTORCH_TESTS_MAX_CHUNK_SIZE)
        ds.image2.extend(np.array([i * np.ones((12, 12)) for i in range(16)]))

    if isinstance(get_base_storage(ds.storage), MemoryProvider):
        with pytest.raises(DatasetUnsupportedPytorch):
            dl = ds.pytorch(num_workers=0)
        return

    dl = ds.pytorch(num_workers=2, transform=to_tuple, batch_size=1)

    for _ in range(2):
        for i, batch in enumerate(dl):
            actual_image = batch[0].numpy()
            expected_image = i * np.ones((1, i + 1, i + 1))
            actual_image2 = batch[1].numpy()
            expected_image2 = i * np.ones((1, 12, 12))
            np.testing.assert_array_equal(actual_image, expected_image)
            np.testing.assert_array_equal(actual_image2, expected_image2)

    dls = ds.pytorch(num_workers=0, transform=to_tuple, batch_size=1, shuffle=True)

    for _ in range(2):
        all_values = []
        for i, batch in enumerate(dls):
            actual_image = batch[0].numpy()
            actual_image2 = batch[1].numpy()

            value = actual_image[0][0][0]
            value2 = actual_image2[0][0][0]
            assert value == value2
            all_values.append(value)

            expected_image = value * np.ones(actual_image.shape)
            expected_image2 = value * np.ones(actual_image2.shape)
            np.testing.assert_array_equal(actual_image, expected_image)
            np.testing.assert_array_equal(actual_image2, expected_image2)

        assert set(all_values) == set(range(16))


@requires_torch
@enabled_datasets
def test_pytorch_with_compression(ds: Dataset):
    # TODO: chunk-wise compression for labels (right now they are uncompressed)
    with ds:
        images = ds.create_tensor(
            "images",
            htype="image",
            sample_compression="png",
            max_chunk_size=PYTORCH_TESTS_MAX_CHUNK_SIZE,
        )
        labels = ds.create_tensor(
            "labels", htype="class_label", max_chunk_size=PYTORCH_TESTS_MAX_CHUNK_SIZE
        )

        assert images.meta.sample_compression == "png"

        images.extend(np.ones((16, 12, 12, 3), dtype="uint8"))
        labels.extend(np.ones((16, 1), dtype="uint32"))

    if isinstance(get_base_storage(ds.storage), MemoryProvider):
        with pytest.raises(DatasetUnsupportedPytorch):
            dl = ds.pytorch(num_workers=0)
        return

    dl = ds.pytorch(num_workers=0, batch_size=1)
    dls = ds.pytorch(num_workers=0, batch_size=1, shuffle=True)

    for dataloader in [dl, dls]:
        for _ in range(2):
            for batch in dataloader:
                X = batch["images"].numpy()
                T = batch["labels"].numpy()
                assert X.shape == (1, 12, 12, 3)
                assert T.shape == (1, 1)


@requires_torch
@enabled_datasets
def test_pytorch_small_old(ds):
    with ds:
        ds.create_tensor("image", max_chunk_size=PYTORCH_TESTS_MAX_CHUNK_SIZE)
        ds.image.extend(np.array([i * np.ones((10, 10)) for i in range(256)]))
        ds.create_tensor("image2", max_chunk_size=PYTORCH_TESTS_MAX_CHUNK_SIZE)
        ds.image2.extend(np.array([i * np.ones((12, 12)) for i in range(256)]))

    if isinstance(get_base_storage(ds.storage), MemoryProvider):
        with pytest.raises(DatasetUnsupportedPytorch):
            dl = dataset_to_pytorch(
                ds, num_workers=0, batch_size=1, python_version_warning=False
            )
        return

    # .pytorch will automatically switch depending on version, this syntax is being used to ensure testing of old code on Python 3.8
    dl = dataset_to_pytorch(
        ds, num_workers=0, batch_size=1, python_version_warning=False
    )

    for _ in range(2):
        for i, batch in enumerate(dl):
            np.testing.assert_array_equal(
                batch["image"].numpy(), i * np.ones((1, 10, 10))
            )
            np.testing.assert_array_equal(
                batch["image2"].numpy(), i * np.ones((1, 12, 12))
            )


@requires_torch
@enabled_datasets
def test_custom_tensor_order(ds):
    with ds:
        tensors = ["a", "b", "c", "d"]
        for t in tensors:
            ds.create_tensor(t, max_chunk_size=PYTORCH_TESTS_MAX_CHUNK_SIZE)
            ds[t].extend(np.random.random((3, 4, 5)))

    if isinstance(get_base_storage(ds.storage), MemoryProvider):
        with pytest.raises(DatasetUnsupportedPytorch):
            dl = dataset_to_pytorch(
                ds, num_workers=0, tensors=["c", "d", "a"], python_version_warning=False
            )
        return

    with pytest.raises(TensorDoesNotExistError):
        dl = ds.pytorch(num_workers=0, tensors=["c", "d", "e"])
    with pytest.raises(TensorDoesNotExistError):
        dl = dataset_to_pytorch(
            ds, num_workers=0, tensors=["c", "e"], python_version_warning=False
        )

    dl_new = ds.pytorch(num_workers=0, tensors=["c", "d", "a"])
    dl_old = dataset_to_pytorch(
        ds, num_workers=0, tensors=["c", "d", "a"], python_version_warning=False
    )

    for dl in [dl_new, dl_old]:
        for i, batch in enumerate(dl):
            c1, d1, a1 = batch
            a2 = batch["a"]
            c2 = batch["c"]
            d2 = batch["d"]
            assert "b" not in batch
            np.testing.assert_array_equal(a1, a2)
            np.testing.assert_array_equal(c1, c2)
            np.testing.assert_array_equal(d1, d2)
            np.testing.assert_array_equal(a1[0], ds.a.numpy()[i])
            np.testing.assert_array_equal(c1[0], ds.c.numpy()[i])
            np.testing.assert_array_equal(d1[0], ds.d.numpy()[i])
            batch = pickle.loads(pickle.dumps(batch))
            c1, d1, a1 = batch
            a2 = batch["a"]
            c2 = batch["c"]
            d2 = batch["d"]
            np.testing.assert_array_equal(a1, a2)
            np.testing.assert_array_equal(c1, c2)
            np.testing.assert_array_equal(d1, d2)
            np.testing.assert_array_equal(a1[0], ds.a.numpy()[i])
            np.testing.assert_array_equal(c1[0], ds.c.numpy()[i])
            np.testing.assert_array_equal(d1[0], ds.d.numpy()[i])

    dls = ds.pytorch(num_workers=0, tensors=["c", "d", "a"])
    for i, batch in enumerate(dls):
        c1, d1, a1 = batch
        a2 = batch["a"]
        c2 = batch["c"]
        d2 = batch["d"]
        assert "b" not in batch
        np.testing.assert_array_equal(a1, a2)
        np.testing.assert_array_equal(c1, c2)
        np.testing.assert_array_equal(d1, d2)
        batch = pickle.loads(pickle.dumps(batch))
        c1, d1, a1 = batch
        a2 = batch["a"]
        c2 = batch["c"]
        d2 = batch["d"]
        np.testing.assert_array_equal(a1, a2)
        np.testing.assert_array_equal(c1, c2)
        np.testing.assert_array_equal(d1, d2)


@requires_torch
def test_readonly_with_two_workers(local_ds):
    local_ds.create_tensor("images", max_chunk_size=PYTORCH_TESTS_MAX_CHUNK_SIZE)
    local_ds.create_tensor("labels", max_chunk_size=PYTORCH_TESTS_MAX_CHUNK_SIZE)
    local_ds.images.extend(np.ones((10, 12, 12)))
    local_ds.labels.extend(np.ones(10))

    base_storage = get_base_storage(local_ds.storage)
    base_storage.flush()
    base_storage.enable_readonly()
    ds = Dataset(storage=local_ds.storage, read_only=True, verbose=False)

    ptds = ds.pytorch(num_workers=2)
    # no need to check input, only care that readonly works
    for _ in ptds:
        pass

    ptds = dataset_to_pytorch(ds)

    for _ in ptds:
        pass


@requires_torch
def test_corrupt_dataset(local_ds, corrupt_image_paths, compressed_image_paths):
    img_good = hub.read(compressed_image_paths["jpeg"][0])
    img_bad = hub.read(corrupt_image_paths["jpeg"])
    with local_ds:
        local_ds.create_tensor("image", htype="image", sample_compression="jpeg")
        for i in range(3):
            for i in range(10):
                local_ds.image.append(img_good)
            local_ds.image.append(img_bad)
    num_samples = 0
    num_batches = 0
    dl = local_ds.pytorch(num_workers=0, batch_size=2)
    for (batch,) in dl:
        num_batches += 1
        num_samples += len(batch)
    assert num_samples == 30
    assert num_batches == 15


@requires_torch
@enabled_datasets
def test_pytorch_local_cache(ds):
    with ds:
        ds.create_tensor("image", max_chunk_size=PYTORCH_TESTS_MAX_CHUNK_SIZE)
        ds.image.extend(([i * np.ones((i + 1, i + 1)) for i in range(16)]))
        ds.create_tensor("image2", max_chunk_size=PYTORCH_TESTS_MAX_CHUNK_SIZE)
        ds.image2.extend(np.array([i * np.ones((12, 12)) for i in range(16)]))

    if isinstance(get_base_storage(ds.storage), MemoryProvider):
        with pytest.raises(DatasetUnsupportedPytorch):
            dl = ds.pytorch(num_workers=2)
        return

    epochs = 2

    for _ in range(epochs):
        dl = ds.pytorch(num_workers=1, batch_size=1, use_local_cache=True)
        for i, batch in enumerate(dl):
            np.testing.assert_array_equal(
                batch["image"].numpy(), i * np.ones((1, i + 1, i + 1))
            )
            np.testing.assert_array_equal(
                batch["image2"].numpy(), i * np.ones((1, 12, 12))
            )

        dls = ds.pytorch(
            num_workers=2,
            batch_size=1,
            shuffle=True,
            use_local_cache=True,
        )
        pytorch_small_shuffle_helper(0, 16, dls)


@requires_torch
def test_groups(local_ds, compressed_image_paths):
    img1 = hub.read(compressed_image_paths["jpeg"][0])
    img2 = hub.read(compressed_image_paths["png"][0])
    with local_ds:
        local_ds.create_tensor(
            "images/jpegs/cats", htype="image", sample_compression="jpeg"
        )
        local_ds.create_tensor(
            "images/pngs/flowers", htype="image", sample_compression="png"
        )
        for _ in range(10):
            local_ds.images.jpegs.cats.append(img1)
            local_ds.images.pngs.flowers.append(img2)
    dl = local_ds.pytorch()
    for cat, flower in dl:
        np.testing.assert_array_equal(cat[0], img1.array)
        np.testing.assert_array_equal(flower[0], img2.array)

    with local_ds:
        local_ds.create_tensor(
            "arrays/x",
        )
        local_ds.create_tensor(
            "arrays/y",
        )
        for _ in range(10):
            local_ds.arrays.x.append(np.random.random((2, 3)))
            local_ds.arrays.y.append(np.random.random((4, 5)))

    dl = local_ds.images.pytorch()
    for cat, flower in dl:
        np.testing.assert_array_equal(cat[0], img1.array)
        np.testing.assert_array_equal(flower[0], img2.array)


@requires_torch
def test_string_tensors(local_ds):
    with local_ds:
        local_ds.create_tensor("strings", htype="text")
        local_ds.strings.extend([f"string{idx}" for idx in range(5)])

    ptds = local_ds.pytorch(batch_size=1)
    for idx, batch in enumerate(ptds):
        np.testing.assert_array_equal(batch["strings"], [f"string{idx}"])

    ptds2 = local_ds.pytorch(batch_size=None)
    for idx, batch in enumerate(ptds2):
        np.testing.assert_array_equal(batch["strings"], f"string{idx}")



import pytest
from hub.util.exceptions import ReadOnlyModeError
from hub.tests.storage_fixtures import enabled_storages


@pytest.mark.xfail(raises=ReadOnlyModeError, strict=True)
@enabled_storages
def test_readonly_del(storage):
    storage.enable_readonly()
    del storage["help!"]


@pytest.mark.xfail(raises=ReadOnlyModeError, strict=True)
@enabled_storages
def test_readonly_set(storage):
    storage.enable_readonly()
    storage["help! im stuck in a storage provider!"] = "salvation"


@pytest.mark.xfail(raises=ReadOnlyModeError, strict=True)
@enabled_storages
def test_readonly_clear(storage):
    storage.enable_readonly()
    storage.clear()


@pytest.mark.xfail(raises=ReadOnlyModeError, strict=True)
@enabled_storages
def test_readonly_flush(storage):
    storage.enable_readonly()
    storage.flush()
